package me.hammerle.code;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map.Entry;
import java.util.Stack;
import java.util.TreeSet;
import me.hammerle.exceptions.PrescriptException;
import me.hammerle.exceptions.NoSuchMethodException;

public class Code implements Comparable<Code>
{
    protected final int line;
    private int subline;
    
    private final String function;
    private final int level;
    private final int pars;
    
    private final Object value;
    private int jump;
    
    public Code(String function, int level, int pars, int line, int subline, Object value, int jump)
    {
        if(function != null)
        {
            this.function = function.toLowerCase();
        }
        else
        {
            this.function = null;
        }
        this.level = level;
        this.pars = pars;
        this.line = line;
        this.subline = subline;
        this.value = value;
        this.jump = jump;
    }
    
    public Code(String function, int level, int pars, int line, int subline)
    {
        this(function.trim(), level, pars, line, subline, null, 0);
    }
    
    public Code(int level, int pars, int line, int subline, Object value)
    {
        this(null, level, pars, line, subline, value, 0);
    }
    
    public void executeFunction(SnuviParser parser, Script sc, Stack<Object> stack)
    {
        //System.out.println("Executing: " + this.toString());
        if(value != null)
        {
            if(value.getClass() == Variable.class)
            {
                stack.push(sc.getVar(((Variable) value).getName()));
                return;
            }
            stack.push(value);
            return;
        }
        Object[] input;
        if(pars > 0)
        {
            input = new Object[pars];
            for(int i = 0; i < pars; i++)
            {
                input[i] = stack.pop();
            }
        }
        else
        {
            input = new Object[0];
        }
        Object output = parser.parseFunction(sc, function, input);
        if(output != Void.TYPE)
        {
            stack.push(output);
        }
    }
    
    public String getFunction()
    {
        return function;
    }

    @Override
    public String toString() 
    {
        StringBuilder sb = new StringBuilder(">>>");
        for(int i = 1; i < level; i++)
        {
            sb.append(">");
        }
        sb.append(" (");
        sb.append(line);
        sb.append(",");
        sb.append(subline);
        sb.append(") ");
        
        if(value != null)
        {
            sb.append("push ");
            sb.append(value.getClass());
            sb.append(" value = '");
            sb.append(value);
            sb.append("'");
            return sb.toString();
        }
        
        sb.append("function ");
        sb.append(function);
        sb.append(" (");
        for(int i = 0; i < pars; i++)
        {
            sb.append("X,");
        }
        if(pars > 0)
        {
            sb.deleteCharAt(sb.length() - 1);
        }
        sb.append(")");
        if(jump != 0)
        {
            sb.append(" (");
            sb.append(jump);
            sb.append(")");
        }
        return sb.toString();
    }
    
    @Override
    public int compareTo(Code o) 
    {
        int i = Integer.compare(line, o.line);
        if(i == 0)
        {
            return Integer.compare(o.subline, subline);
        }
        return i;
    }
    
    public int getJumpLine()
    {
        return jump;
    }
    
    // -----------------------------------------------------------------------------------
    // code builder
    // -----------------------------------------------------------------------------------
    
    private static SnuviParser parser = null;
    private static String scriptName = null;
    private static int sublines = 0;
    
    private static int findEndOfLine(String code, int pos)
    {
        int start = pos;
        int length = code.length();
        char c;
        boolean text = false;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            switch(c)
            {
                case ';':
                case '{':
                case '}':
                    return pos;
                case '"':
                    text = !text;
                    break;
            }
            pos++;
        }
        return start;
    }
    
    private static int findStartOfSyntax(String code, int pos)
    {
        int bracketCounter = 0;
        char c;
        boolean text = false;
        while(pos >= 0 && code.charAt(pos) == ' ')
        {
            pos--;
        }
        while(pos >= 0)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos--;
                continue;
            }
            switch(c)
            {
                case ';':
                case '{':
                case '}':
                    if(bracketCounter != 0)
                    {
                        throw new PrescriptException(scriptName, code.substring(pos - 1, Math.min(code.length(), pos + 20)), "unbalanced ()"); 
                    }
                    return pos + 1;
                case ' ':
                case '+':
                case '-':
                case '*':
                case '/':
                case '^':
                case '=':
                case '>':
                case '<':
                case '!':
                case '%':
                case ',':
                case '&':
                case '|':
                    if(bracketCounter != 0)
                    {
                        break;
                    }
                    return pos + 1;
                case '"':
                    text = !text;
                    break;
                case ')':
                    bracketCounter++;
                    break;
                case '(':
                    if(bracketCounter == 0)
                    {
                        return pos + 1;
                    }
                    bracketCounter--;
                    break;
            }
            pos--;
        }
        return 0;
    }
    
    private static int findEndOfSyntax(String code, int pos)
    {
        int bracketCounter = 0;
        char c;
        boolean text = false;
        while(pos < code.length() && code.charAt(pos) == ' ')
        {
            pos++;
        }
        while(pos < code.length())
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            switch(c)
            {
                case ';':
                case '{':
                case '}':
                    if(bracketCounter != 0)
                    {
                        throw new PrescriptException(scriptName, code.substring(pos - 1, Math.min(code.length(), pos + 20)), "unbalanced ()"); 
                    }
                    return pos;
                case ' ':
                case '+':
                case '-':
                case '*':
                case '/':
                case '^':
                case '=':
                case '>':
                case '<':
                case '!':
                case '%':
                case ',':
                case '&':
                case '|':
                    if(bracketCounter != 0)
                    {
                        break;
                    }
                    return pos;
                case '"':
                    text = !text;
                    break;
                case '(':
                    bracketCounter++;
                    break;
                case ')':
                    if(bracketCounter == 0)
                    {
                        return pos;
                    }
                    bracketCounter--;
                    break;
            }
            pos++;
        }
        return code.length();
    }
    
    private static int findSyntax(String find, String code, int pos)
    {
        int length = code.length();
        char c;
        boolean text = false;
        String s;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            if(c == '"')
            {
                text = !text;
                pos++;
                continue;
            }
            s = code.substring(pos);
            // additionel check for e.g. difference between + and +=
            if(s.startsWith(find) && !s.startsWith(find + "=") && code.charAt(pos - 1) != '=') 
            {
                return pos;
            }
            pos++;
        }
        return -1;
    }
    
    private static void changeBiSyntax(String syntax, String f, StringBuilder sb, boolean b)
    {
        int pos = -1;
        int end;
        int start;
        int slength = syntax.length();
        StringBuilder newSyntax;
        while(pos < sb.length())
        {
            pos = findSyntax(syntax, sb.toString(), pos + 1);
            if(pos == -1)
            {
                break;
            }
            start = findStartOfSyntax(sb.toString(), pos - 1);
            end = findEndOfSyntax(sb.toString(), pos + slength);
            newSyntax = new StringBuilder(f);
            newSyntax.append("(");
            if(b)
            {
                newSyntax.append("\"");
            }
            newSyntax.append(sb.substring(start, pos).trim());
            if(b)
            {
                newSyntax.append("\"");
            }
            if(end > pos)
            {
                newSyntax.append(",");
                newSyntax.append(sb.substring(pos + slength, end).trim());
            }
            newSyntax.append(")");
            pos -= end - start - newSyntax.length();
            //System.out.println(sb.substring(start, end) + "   ===>    " + newSyntax);
            sb.replace(start, end, newSyntax.toString());
        }
    }
    
    private static void changeUnSyntax(String syntax, String f, StringBuilder sb)
    {
        int pos = -1;
        int end;
        int start;
        int slength = syntax.length();
        String first;
        StringBuilder newSyntax;
        while(pos < sb.length())
        {
            pos = findSyntax(syntax, sb.toString(), pos + 1);
            if(pos == -1)
            {
                break;
            }
            start = findStartOfSyntax(sb.toString(), pos - 1);
            end = findEndOfSyntax(sb.toString(), pos + slength);
            newSyntax = new StringBuilder("setvar(\"");
            first = sb.substring(start, pos).trim();
            newSyntax.append(first);
            newSyntax.append("\",");
            newSyntax.append(f);
            newSyntax.append("(");
            newSyntax.append(first);
            newSyntax.append(",");
            newSyntax.append(sb.substring(pos + slength, end).trim());
            newSyntax.append("))");
            pos -= end - start - newSyntax.length();
            sb.replace(start, end, newSyntax.toString());
        }
    }
    
    private static String replaceChar(char find, char replacement, String code)
    {
        char[] chars = code.toCharArray();
        int length = code.length();
        char c;
        boolean text = false;
        int pos = 0;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            if(c == '"')
            {
                text = !text;
                pos++;
                continue;
            }
            if(c == find)
            {
                chars[pos] = replacement;
            }
            pos++;
        }
        return new String(chars);
    }
    
    private static void addKeyWordBrackets(String keyWord, StringBuilder sb)
    {
        String with = keyWord + "()";
        int length = keyWord.length();
        int pos = 0;
        while(true)
        {
            pos = findSyntax(keyWord, sb.toString(), pos);
            if(pos == -1)
            {
                break;
            }
            if(!sb.substring(pos).startsWith(with))
            {
                sb.replace(pos, pos + length, with);
            }
            pos += length;
        }
    }
    
    public static Code[] generate(SnuviParser parser, String scriptName, String code, HashMap<String, Integer> gotos)
    {
        Code.scriptName = scriptName;
        Code.parser = parser;
        // comments
        int old = 0;
        int pos;
        StringBuilder sb = new StringBuilder(code);
        while(true)
        {
            old = findSyntax("/*", sb.toString(), old);
            if(old == -1)
            {
                break;
            }
            pos = findSyntax("*/", sb.toString(), old);
            if(pos == -1)
            {
                throw new PrescriptException(scriptName, sb.substring(old, Math.min(old + 20, sb.length())), "/* without */");
            }
            sb.replace(old, pos + 2, "");
        }
        old = 0;
        while(true)
        {
            old = findSyntax("//", sb.toString(), old);
            if(old == -1)
            {
                break;
            }
            pos = findSyntax("\n", sb.toString(), old);
            if(pos == -1)
            {
                sb.replace(old, sb.length(), "");
                break;
            }
            sb.replace(old, pos + 1, "");
        }
        // end comments

        // key words without ()
        addKeyWordBrackets("else", sb);        
        addKeyWordBrackets("try", sb);   
        addKeyWordBrackets("catch", sb);   
        // end key words

        // replacing of function syntax
        changeBiSyntax("^", "math.pow", sb, false);
        changeBiSyntax("/", "div", sb, false);
        changeBiSyntax("%", "math.mod", sb, false);
        changeBiSyntax("*", "mul", sb, false);
        changeBiSyntax("-", "sub", sb, false);
        changeBiSyntax("+", "add", sb, false);

        changeBiSyntax("==", "equal", sb, false);
        changeBiSyntax("!=", "notequal", sb, false);
        changeBiSyntax("<", "less", sb, false);
        changeBiSyntax(">", "greater", sb, false);
        changeBiSyntax(">=", "greaterequal", sb, false);
        changeBiSyntax("<=", "lessequal", sb, false);

        changeBiSyntax("||", "or", sb, false);
        changeBiSyntax("&&", "and", sb, false);

        changeUnSyntax("+=", "add", sb);
        changeUnSyntax("-=", "sub", sb);
        changeUnSyntax("/=", "div", sb);
        changeUnSyntax("*=", "mul", sb);

        changeBiSyntax("=", "setvar", sb, true);

        code = sb.toString();
        // end of substitution
        // split code into lines
        ArrayList<Code> list = new ArrayList<>();
        sublines = 0;
        String actual;
        int length = code.length();
        int level = 1;
        pos = 0;
        int line = 0;
        while(pos < length)
        {
            old = pos;
            pos = findEndOfLine(code, pos);
            actual = code.substring(old, pos).trim();
            if(actual.startsWith("@"))
            {
                // sets the right layer of the goto, the exact position is set later
                if(gotos.put(actual.substring(1), line) != null)
                {
                    throw new PrescriptException(scriptName, code.substring(old, Math.min(code.length(), pos + 20)), "double goto");
                }
                pos++;
                continue;
            }
            switch(code.charAt(pos))
            {
                case '{':
                    line++;
                    splitFunctions(list, actual, level, line);
                    level++;
                    break;
                case '}': 
                    level--;
                    break;
                case ';': 
                    line++;
                    splitFunctions(list, actual, level, line);
                    break;
            }
            pos++;
        }
        // end of code splitting

        // tree for sorting and inserting of code
        TreeSet<Code> tree = new TreeSet(list);
        Code[] c = tree.toArray(new Code[tree.size()]);

        // insert "gotoline" at while
        String function;
        int baseLevel;
        boolean lastLineChecker;
        for(int i = 0; i < c.length; i++)
        {
            function = c[i].function;
            if(function == null || !function.equals("while"))
            {
                continue;
            }
            baseLevel = c[i].level;
            // keeping sure "gotoline" is inserted at least at the end of the code
            lastLineChecker = true;
            for(int j = i + 1; j < c.length; j++)
            {
                if(c[j].level <= baseLevel)
                {
                    Code gotoLine = new Code("gotoline", c[i].level, 0, c[j].line, c[j].subline + 1);
                    while(tree.contains(gotoLine)) // prevent whiles from overwriting other whiles
                    {
                        gotoLine.subline++;
                    }
                    tree.add(gotoLine);
                    lastLineChecker = false;
                    break;
                }
            }
            if(lastLineChecker)
            {
                sublines++;
                Code gotoLine = new Code("gotoline", c[i].level, 0, c[c.length - 1].line + 1, sublines);
                while(tree.contains(gotoLine)) // prevent whiles from overwriting other whiles
                {
                    gotoLine.subline++;
                }
                tree.add(gotoLine);
            }
        }
        // end of while handling
        // generating gotos of key words
        c = tree.toArray(new Code[tree.size()]);

        boolean whileCheck = false;
        int lastLineChange = 0;
        line = 0;
        for(int i = 0; i < c.length; i++)
        {
            function = c[i].function;
            if(c[i].line != line)
            {
                line = c[i].line;
                lastLineChange = i;
            }
            if(function == null)
            {
                continue;
            }
            switch(function)
            {
                case "if":
                case "else":
                case "try":
                case "catch":
                    break;
                case "while":
                    whileCheck = true;
                    break;
                default:
                    continue;
            }
            baseLevel = c[i].level;
            for(int j = i + 1; j < c.length; j++)
            {
                if(c[j].level <= baseLevel)
                {
                    c[i].jump = j - i; // key word pos - end pos = jump
                    if(whileCheck)
                    {
                        // setting right jump for gotoline of while
                        whileCheck = false;
                        c[j].jump = lastLineChange - j;
                    }
                    break;
                }
            }
            if(c[i].jump == 0)
            {
               c[i].jump = tree.size() - i; 
            }
        }
        // end of key word jumps
        // correct wrong gotos
        for(Entry<String, Integer> i : gotos.entrySet())
        {
            int value = i.getValue() + 1;
            for(int j = value - 1; j < c.length; j++)
            {
                if(c[j].line == value)
                {
                    i.setValue(j);
                    break;
                }
            }
        }
        // end
        //java.util.Arrays.stream(c).forEach(co -> System.out.println(co.toString()));
        //gotos.forEach((k, v) -> System.out.println(k + "   " + v));
        //System.exit(0);
        return c;
    } 
    
    private static int findOpenBracket(String code, int pos)
    {
        int length = code.length();
        char c;
        boolean text = false;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            switch(c)
            {
                case '(':
                    return pos;
                case '"':
                    text = !text;
                    break;
            }
            pos++;
        }
        return -1;
    }
    
    private static int findClosingBracket(String code, int pos)
    {
        if(pos == -1)
        {
           pos++;
        }
        int length = code.length();
        char c;
        boolean text = false;
        int brackets = 0;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            switch(c)
            {
                case '(':
                    brackets++;
                    break;
                case ')':
                    brackets--;
                    if(brackets == 0)
                    {
                        return pos;
                    }
                    break;
                case '"':
                    text = !text;
                    break;
            }
            pos++;
        }
        return -1;
    }
    
    private static ArrayList<String> splitComma(String code)
    {
        ArrayList<String> list = new ArrayList<>();
        int length = code.length();
        int pos = 0;
        int old = 0;
        int brackets = 0;
        char c;
        boolean text = false;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            
            switch(c)
            {
                case '"':
                    text = !text;
                    break;
                case '(':
                    brackets++;
                    break;
                case ')':
                    brackets--;
                    break;
                case ',':
                    if(brackets == 0)
                    {
                        list.add(code.substring(old, pos));
                        old = pos + 1;
                    }
                    break;
            }
            pos++;
        }
        if(old < pos)
        {
            list.add(code.substring(old, pos));
        }
        return list;
    }
    
    private static void splitFunctions(ArrayList<Code> list, String f, int level, int line)
    {
        sublines++;
        int start = findOpenBracket(f, 0);
        int end = findClosingBracket(f, start);
        if((start != -1 || end != -1) && (((start == -1 || end == -1) && start != end) || end != f.length() - 1))
        {
            throw new PrescriptException(scriptName, "unbalanced ()", f);
        }
        if(start == -1)
        {
            list.add(new Code(level, 0, line, sublines, convertInput(f)));
            return;
        }
        else if(start >= end - 1)
        {
            String functionName = f.substring(0, start).trim();
            if(!parser.isRegisteredFunction(functionName))
            {
                throw new NoSuchMethodException(scriptName, f, functionName);
            }
            list.add(new Code(functionName, level, 0, line, sublines));
            return;
        }
        ArrayList<String> splitted = splitComma(f.substring(start + 1, end));
        String functionName = f.substring(0, start).trim();
        if(!parser.isRegisteredFunction(functionName))
        {
            throw new NoSuchMethodException(scriptName, f, functionName);
        }
        list.add(new Code(functionName, level, splitted.size(), line, sublines));
        splitted.forEach(s -> splitFunctions(list, s, level, line));
    }
    
    public static Object convertInput(String s)
    {
        if(s == null)
        {
            return null;
        }
        s = s.trim();
        if(s.length() == 0)
        {
            return "";
        }  
        if(s.charAt(0) == '"' && s.charAt(s.length() - 1) == '"')
        {
            if(s.length() == 1)
            {
                return "\"";
            }
            else if(s.charAt(1) == '$')
            {
                return s.substring(2, s.length() - 1);
            }
            return s.substring(1, s.length() - 1);
        }
        else if(s.equals("true"))
        {
            return true;
        }
        else if(s.equals("false"))
        {
            return false;
        }
        else if(s.equals("null"))
        {
            return null;
        }
        try
        {
            return Double.parseDouble(s);
        }
        catch(NumberFormatException ex)
        {
            return new Variable(s);
        }
    }
}
